/*  
 * drivers/cpufreq/cpufreq_stresstest.c
 *
 *  Copyright (C)  2001 Russell King
 *            (C)  2003 Venkatesh Pallipadi <venkatesh.pallipadi@intel.com>.
 *                      Jun Nakajima <jun.nakajima@intel.com>
 *            (C)  2009 Dell Inc (Steven_Finney@dell.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This is a hacked version of cpufreq_ondemand to test the behavior of the 
 * hardware and lower-level DVFS code under aggressive frequency change. It
 * is a replacement governor that will force frequency changes at a fixed
 * (settable) interval regardless of system load; there are various ways to
 * parameterize the frequency sequence. The intent is to characterize the
 * behavior of the hardware and clock/PLL code under difficult conditions.
 *
 * Caveats: most multi-cpu support (enable, etc) has been (sloppily) 
 * removed (I didn't understand how it worked, couldn't test it, and it got
 * in the way). This is a TEST module only and has not been bulletproofed
 * the way production code should be. Changes in policy are ignored; 
 * policy in general is ignored. The provided frequency table is taken as 
 * Truth. scaling_min/max_freq is ignored.
 *
 * This  driver currently is not built into the Kconfig system; if you want 
 * use it, the simplest thing to do is to add it to cpufreq/Makefile.
 *
 *
 * Usage notes:
 *	FS=/sys/devices/system/cpu/cpu0/cpufreq
 *	ST=$FS/stresstest
 *
 * To enable:
 *	echo stresstest > $FS/scaling_governor 
 *	(The default startup mode will run at a fixed 998400 (KHz)).  
 *
 * $FS/stresstest files/parameters:
 *	available_{modes,frequencies} shows the supported 
 *		modes/frequencies (KHz) (RO).
 *	latency: time between frequency changes (ms)(RW)
 *	mode: frequency sequence mode, see below. RW
 *	{min,max}: freq in KHz, see below. RW
 *	print_freq: set to 1 to printout all frequency changes. RW.
 *	
 *	Modes:
 *		fixed: run statically at the frequency 'max'.
 *		dual: switch back and forth  between frequency 'max' and 'min'
 *		random: each switch is a random choice of the available
 *			frequencies. Repetitions may occur.
 *		exhaustive: go through a sequence that exercises every possible
 *			pairwise transition. This is not a minimal sequence,
 *			but it's a simple imiplementation.
 *		list: use a hard-coded sequence; the sequence is currently
 *			one that crashes Surf (3145 build).
 *	
 *     Example: Cycle between 245 Mhz and 984 MHz at a 100 ms interval.
 *		echo 100 > $ST/latency; echo 245000 > $ST/min 
 *		echo 998400 > $ST/max; echo dual > $ST/mode
 *
 * Qualcomm Snapdragon specifics: the frequencies in stress_tuners are
 * based on Snapdragon. The frequency table used is constructed from 
 * enabled ("1") entries in acpu_freq_tbl[] (acpuclock-8x50.c); you'll need
 * to edit that file if you want to test frequencies not currently enabled
 * by Qualcomm. 
 *
 * sfinney, 6/19/09, 8/25/09
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/cpufreq.h>
#include <linux/cpu.h>
#include <linux/jiffies.h>
#include <linux/kernel_stat.h>
#include <linux/mutex.h>
#include <linux/random.h>

#define DRIVER_NAME	"cpufreq_stresstest"
#define DRIVER_NAME_SHORT	"stresstest"

static void do_stress_timer(struct work_struct *work);

struct cpu_stress_info_s {
	struct cpufreq_policy *cur_policy;
 	struct delayed_work work;
	struct cpufreq_frequency_table *freq_table;
	int cpu;	
} cpu_stress_info;



static DEFINE_MUTEX(stress_mutex);

static struct workqueue_struct	*kstresstest_wq;




enum {ST_RANDOM, ST_EXHAUSTIVE, ST_DUAL, ST_FIXED, ST_LIST} ;

struct mode {
	int mode;
	char name[CPUFREQ_NAME_LEN];
} modes[] = 
{ 	{ST_RANDOM, 	"random"},
 	{ST_EXHAUSTIVE, "exhaustive"},
 	{ST_DUAL, 	"dual"},
 	{ST_FIXED, 	"fixed"},
 	{ST_LIST, 	"list"},
	{-1},
};


static struct stress_tuners {
	
	unsigned int interval;	
	
	unsigned int min;
	unsigned int max;
	unsigned int mode;	
 	unsigned int print_freq;  
} stress_tuners_ins = {
	.interval = 1000,	
	.min = 576000,		
	.max = 998400 ,		
	.mode = ST_FIXED,
	.print_freq = 0,
};



#define AVAILABLE_FREQ_MAX  50	
unsigned int available_freq_list[AVAILABLE_FREQ_MAX];
unsigned int available_freq_cnt = 0;



unsigned int static_freq_list[] = {768000,  768000, 245000, 384000};
unsigned int static_freq_cnt = sizeof(static_freq_list)/sizeof(unsigned int); 	


#define MAX_CYCLE	20
unsigned int freq_list[MAX_CYCLE + 1];
unsigned int freq_list_cnt;	
unsigned int freq_list_index;


void freq_list_init(void)
{
	int i;
	switch(stress_tuners_ins.mode)
	{
		
		case ST_DUAL:
			freq_list[0] = stress_tuners_ins.max;
			freq_list[1] = stress_tuners_ins.min;
			freq_list_cnt = 2;
			break;

		case ST_LIST:
			for (i = 0; i < static_freq_cnt; i++)
				freq_list[i] = static_freq_list[i];
			freq_list_cnt = static_freq_cnt;
			break;

		case ST_EXHAUSTIVE:
		case ST_FIXED:
		case ST_RANDOM:
			break;	
			
		default:
			printk("%s: unknown mode %d\n",
				DRIVER_NAME, stress_tuners_ins.mode);
			break;	
	};

	
	freq_list_index = 0;	
};



static unsigned int exhaustive_next(void)
{
	static int i = 0, j = 0, use_anchor = 1;
	int next_freq;

	if (use_anchor) {
		use_anchor = 0;
		return available_freq_list[i];
	}
	use_anchor = 1;
	next_freq = available_freq_list[j++];
	if (j == available_freq_cnt) {
		j = 0;
		if (++i == available_freq_cnt)
			i = 0;
	}
	return next_freq;
}

static unsigned int get_next_freq(void)
{
	unsigned int next_freq = 0; 
	switch(stress_tuners_ins.mode)
	{

		case ST_FIXED:
			next_freq = stress_tuners_ins.max;
			break;

		case ST_EXHAUSTIVE:
			next_freq = exhaustive_next();
			break;

		case ST_DUAL:
		case ST_LIST:
			next_freq = freq_list[freq_list_index++];
			if (freq_list_index == freq_list_cnt)
				freq_list_index = 0;
			break;

		case ST_RANDOM:
			next_freq = available_freq_list[random32() %
				available_freq_cnt];
			break;

		default:
			
			printk("%s: unknown mode %d\n",
				DRIVER_NAME, stress_tuners_ins.mode);
			break;	
	}

	if (stress_tuners_ins.print_freq)
		printk("%s: next_freq = %d\n", __FUNCTION__, next_freq);

	return next_freq;
}



static char * mode_name(int mode)
{	int i;
	for (i = 0; modes[i].mode != -1; i++) {
		if (modes[i].mode == mode)
			return modes[i].name;
	};
	return "Unknown mode";
}

static int  mode_int(char *name)
{	int i;
	for (i = 0; modes[i].mode != -1; i++) {
		if (!strcmp(name, modes[i].name))
			return modes[i].mode;
	};
	return -1;
}



#define show_one(file_name, object)					\
static ssize_t show_##file_name						\
(struct cpufreq_policy *unused, char *buf)				\
{									\
	return sprintf(buf, "%u\n", stress_tuners_ins.object);		\
}



show_one(interval, interval);
show_one(min, min);
show_one(max, max);
show_one(print_freq, print_freq);

static ssize_t show_mode (struct cpufreq_policy *unused, char *buf)
{
	return sprintf(buf, "%s\n", mode_name(stress_tuners_ins.mode));
}


static ssize_t show_available_modes (struct cpufreq_policy *unused, char *buf)
{
	ssize_t i = 0;
	int j;


	for (j = 0; modes[j].mode != -1; j++) {

		if (i >= (ssize_t) ((PAGE_SIZE / sizeof(char)) - (CPUFREQ_NAME_LEN + 2)))
			goto out;
		i += scnprintf(&buf[i], CPUFREQ_NAME_LEN, "%s ", modes[j].name);
	}
out:
	i += sprintf(&buf[i], "\n");
	return i;
}


#define FREQ_PRINT_SIZE 8
static ssize_t show_available_freqs (struct cpufreq_policy *unused, char *buf)
{
	ssize_t i = 0;
	int j;


	for (j = 0; j < available_freq_cnt; j++) {
		if (i >= (ssize_t) ((PAGE_SIZE / sizeof(char)) - (FREQ_PRINT_SIZE + 2)))
			goto out;
		i += scnprintf(&buf[i], FREQ_PRINT_SIZE, "%d\t", 
			available_freq_list[j]);
		if ((j%8) == 7)
		 	buf[i - 1] = '\n';

	}
out:
	i += sprintf(&buf[i], "\n");
	return i;
}

static ssize_t store_interval(struct cpufreq_policy *unused,
		const char *buf, size_t count)
{
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	mutex_lock(&stress_mutex);
	if (ret != 1) {
		mutex_unlock(&stress_mutex);
		return -EINVAL;
	}

	stress_tuners_ins.interval = input;
	mutex_unlock(&stress_mutex);

	return count;
}


static ssize_t store_min(struct cpufreq_policy *unused,
		const char *buf, size_t count)
{
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	mutex_lock(&stress_mutex);
	if (ret != 1) {
		mutex_unlock(&stress_mutex);
		return -EINVAL;
	}

	stress_tuners_ins.min = input;
	freq_list_init();	
	mutex_unlock(&stress_mutex);

	return count;
}

static ssize_t store_max(struct cpufreq_policy *unused,
		const char *buf, size_t count)
{
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	mutex_lock(&stress_mutex);
	if (ret != 1) {
		mutex_unlock(&stress_mutex);
		return -EINVAL;
	}

	stress_tuners_ins.max = input;
	freq_list_init();
	mutex_unlock(&stress_mutex);

	return count;
}


static ssize_t store_print_freq(struct cpufreq_policy *unused,
		const char *buf, size_t count)
{
	unsigned int input;
	int ret;
	ret = sscanf(buf, "%u", &input);

	mutex_lock(&stress_mutex);
	if (ret != 1) {
		mutex_unlock(&stress_mutex);
		return -EINVAL;
	}

	if (input != 0)
		input = 1;
	stress_tuners_ins.print_freq = input;
	mutex_unlock(&stress_mutex);

	return count;
}

static ssize_t store_mode(struct cpufreq_policy *unused,
		const char *buf, size_t count)
{
	int ret, mode;
	char str_mode[20];

	ret = sscanf(buf, "%19s", str_mode);
	if (ret != 1)
		return -EINVAL;
	mode = mode_int(str_mode);
	if (mode == -1) {
		printk("Cpufreq stresstest governor: unknown mode \"%s\"\n",
			str_mode);
		return -EINVAL;
	};

	mutex_lock(&stress_mutex);
	if (ret != 1) {
		mutex_unlock(&stress_mutex);
		return -EINVAL;
	}

	stress_tuners_ins.mode = mode;
	freq_list_init();
	mutex_unlock(&stress_mutex);

	return count;
}

#define define_one_rw(_name) \
static struct freq_attr _name = \
__ATTR(_name, 0644, show_##_name, store_##_name)

define_one_rw(interval);
define_one_rw(min);
define_one_rw(max);
define_one_rw(mode);
define_one_rw(print_freq);

#define define_one_ro(_name) \
static struct freq_attr _name = \
__ATTR(_name, 0444, show_##_name, NULL)

define_one_ro(available_modes);
define_one_ro(available_freqs);


static struct attribute * stress_attributes[] = {
	&interval.attr,
	&min.attr,
	&max.attr,
	&mode.attr,
	&print_freq.attr,
	&available_modes.attr,
	&available_freqs.attr,
	NULL
};

static struct attribute_group stress_attr_group = {
	.attrs = stress_attributes,
	.name = DRIVER_NAME_SHORT,
};



static void do_stress_timer(struct work_struct *work)
{
	struct cpu_stress_info_s *stress_info =
		container_of(work, struct cpu_stress_info_s, work.work);
	unsigned int cpu = stress_info->cpu;
	unsigned int next_freq;


	if (lock_policy_rwsem_write(cpu) < 0)
		return;

	next_freq = get_next_freq();
	__cpufreq_driver_target(stress_info->cur_policy, next_freq,
	                        	CPUFREQ_RELATION_H);
	queue_delayed_work_on(cpu, kstresstest_wq, &stress_info->work, 
		msecs_to_jiffies(stress_tuners_ins.interval));
	unlock_policy_rwsem_write(cpu);
}

static inline void stress_timer_init(struct cpu_stress_info_s *stress_info)
{

	INIT_DELAYED_WORK(&stress_info->work, do_stress_timer);
	queue_delayed_work_on(stress_info->cpu, kstresstest_wq, 
		&stress_info->work, msecs_to_jiffies(stress_tuners_ins.interval));
}

static inline void stress_timer_exit(struct cpu_stress_info_s *stress_info)
{
	cancel_delayed_work(&stress_info->work);
}

static int cpufreq_governor_stress(struct cpufreq_policy *policy,
				   unsigned int event)
{
	unsigned int cpu = policy->cpu;
	struct cpu_stress_info_s *this_stress_info;
	struct cpufreq_frequency_table *table;
	int rc, i;

	this_stress_info = &cpu_stress_info;

	switch (event) {
	case CPUFREQ_GOV_START:
		if ((!policy->cur))
			return -EINVAL;


		mutex_lock(&stress_mutex);

		rc = sysfs_create_group(&policy->kobj, &stress_attr_group);
		if (rc) {
			mutex_unlock(&stress_mutex);
			return rc;
		}

		
		cpu_stress_info.cur_policy = policy;
		this_stress_info->cpu = cpu;	
		this_stress_info->freq_table = cpufreq_frequency_get_table(cpu);
		
		
		
		available_freq_cnt = 0;
		table = this_stress_info->freq_table;
		for (i = 0; table[i].frequency != CPUFREQ_TABLE_END; i++) {
			if (table[i].frequency != CPUFREQ_ENTRY_INVALID) {
				if (available_freq_cnt >=  AVAILABLE_FREQ_MAX) {
					printk("%s: too many frequencies (%d), truncating\n",
						DRIVER_NAME, available_freq_cnt);
					break;
				};
				available_freq_list[available_freq_cnt++] =
					table[i].frequency;
						
			};
		};
	

		freq_list_init();

		
		stress_timer_init(this_stress_info);

		mutex_unlock(&stress_mutex);
		break;

	case CPUFREQ_GOV_STOP:
		mutex_lock(&stress_mutex);
		stress_timer_exit(this_stress_info);
		sysfs_remove_group(&policy->kobj, &stress_attr_group);
		mutex_unlock(&stress_mutex);

		break;

	case CPUFREQ_GOV_LIMITS:
		mutex_lock(&stress_mutex);
		if (policy->max < this_stress_info->cur_policy->cur)
			__cpufreq_driver_target(this_stress_info->cur_policy,
			                        policy->max,
			                        CPUFREQ_RELATION_H);
		else if (policy->min > this_stress_info->cur_policy->cur)
			__cpufreq_driver_target(this_stress_info->cur_policy,
			                        policy->min,
			                        CPUFREQ_RELATION_L);
		mutex_unlock(&stress_mutex);
		break;
	}
	return 0;
}

struct cpufreq_governor cpufreq_gov_stresstest = {
	.name			= DRIVER_NAME_SHORT,
	.governor		= cpufreq_governor_stress,
	.max_transition_latency = 10 * 1000 * 1000,	
	.owner			= THIS_MODULE,
};
EXPORT_SYMBOL(cpufreq_gov_stresstest);

static int __init cpufreq_gov_stress_init(void)
{
	kstresstest_wq = create_workqueue("kstresstest");
	if (!kstresstest_wq) {
		printk(KERN_ERR "Creation of kstresstest failed\n");
		return -EFAULT;
	}
	return cpufreq_register_governor(&cpufreq_gov_stresstest);
}

static void __exit cpufreq_gov_stress_exit(void)
{
	cpufreq_unregister_governor(&cpufreq_gov_stresstest);
	destroy_workqueue(kstresstest_wq);
}


MODULE_AUTHOR("Venkatesh Pallipadi <venkatesh.pallipadi@intel.com>");
MODULE_AUTHOR("Alexey Starikovskiy <alexey.y.starikovskiy@intel.com>");
MODULE_AUTHOR("Steve Finney");
MODULE_DESCRIPTION("cpufreq_stresstest");
MODULE_LICENSE("GPL");

 
module_init(cpufreq_gov_stress_init);

module_exit(cpufreq_gov_stress_exit);
