/* drivers/usb/function/mtp.c
 *
 * Author: Ben Bostwick
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>

#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>

#include <asm/atomic.h>
#include <asm/uaccess.h>

#include "usb_function.h"
#include <linux/usb/gadgetfs.h>

#define TXN_MAX 16384

#define CTL_WRITE_BUF_SZ 64


#define USB_MTP_IOC_MAGIC 0xFF

#define USB_MTP_FUNC_IOC_CANCEL_REQUEST_SET _IOW(USB_MTP_IOC_MAGIC, 0x20, int)
#define USB_MTP_FUNC_IOC_CANCEL_REQUEST_GET _IOW(USB_MTP_IOC_MAGIC, 0x21, int)
#define USB_MTP_FUNC_IOC_GET_EXTENDED_EVENT_DATA_SET    _IOW(USB_MTP_IOC_MAGIC, 0x22, int)
#define USB_MTP_FUNC_IOC_GET_EXTENDED_EVENT_DATA_GET    _IOW(USB_MTP_IOC_MAGIC, 0x23, int)
#define USB_MTP_FUNC_IOC_DEVICE_RESET_REQUEST_SET   _IOW(USB_MTP_IOC_MAGIC, 0x24, int)
#define USB_MTP_FUNC_IOC_DEVICE_RESET_REQUEST_GET   _IOW(USB_MTP_IOC_MAGIC, 0x25, int)
#define USB_MTP_FUNC_IOC_GET_DEVICE_STATUS_SET  _IOW(USB_MTP_IOC_MAGIC, 0x26, int)
#define USB_MTP_FUNC_IOC_GET_DEVICE_STATUS_GET  _IOW(USB_MTP_IOC_MAGIC, 0x27, int)

#define STRING_MSFT_SECRET      0xEE
#define MSFT_SECRET             "MSFT100"

#define PTP_CANCEL_REQUEST      0x64
#define PTP_RESET               0x66
#define PTP_GET_DEVICE_STATUS   0x67

#define PTP_DATATYPE_UINT16     0x0004

#define PTP_RESPONSE_OK         0x2001
#define PTP_RESPONSE_BUSY       0x2019

#define PTP_RESPONSE_OK_CAP     ((PTP_RESPONSE_OK << 16) | ((PTP_DATATYPE_UINT16)))
#define PTP_RESPONSE_BUSY_CAP    ((PTP_RESPONSE_BUSY << 16) | ((PTP_DATATYPE_UINT16)))

#define MTP_SUPPORT_CANCEL


#define RX_REQ_MAX 4
#define TX_REQ_MAX 4

#define MTP_FUNCTION_NAME "mtp"

enum {
	MSM_MTP_CTL_DEBUG = 1U << 0,
	MSM_MTP_IO_DEBUG = 1U << 1,
	MSM_MTP_ERR = 1U << 2,
	MSM_MTP_INFO = 1U << 3,
};

#define MTP_LOG_INFO
#define MTP_LOG_CTL

#define MTP_LOG_ERR

#ifdef MTP_LOG_INFO
#define DBG_INFO(x...) do {				\
			printk(x);		\
	} while (0)

#else
#define DBG_INFO(x...) do {} while (0)
#endif

#ifdef MTP_LOG_CTL
#define DBG_CTL(x...) do {				\
			printk(x);		\
	} while (0)

#else
#define DBG_CTL(x...) do {} while (0)
#endif

#ifdef MTP_LOG_IO
#define DBG_IO(x...) do {               \
			printk(x);		\
	} while (0)

#else
#define DBG_IO(x...) do {} while (0)
#endif

#ifdef MTP_LOG_ERR
#define DBG_ERR(x...) do {              \
			printk(x);		\
	} while (0)

#else
#define DBG_ERR(x...) do {} while (0)
#endif


struct mtp_context
{
        int online;
        int error;

#ifdef MTP_SUPPORT_CANCEL
        int cancel_req_excl;
        spinlock_t cancel_lock;
#endif

        atomic_t read_excl;
        atomic_t write_excl;
        atomic_t open_excl;
        atomic_t ctl_read_excl;
        atomic_t ctl_write_excl;
        atomic_t ctl_open_excl;
        atomic_t ctl_enable_excl;
        atomic_t enable_excl;
        spinlock_t lock;

        struct usb_endpoint *ctl;
        struct usb_endpoint *out;
        struct usb_endpoint *in;
        struct usb_endpoint *status;

        struct list_head tx_idle;
        struct list_head rx_idle;
        struct list_head rx_done;

        wait_queue_head_t read_wq;
        wait_queue_head_t write_wq;
        wait_queue_head_t cancel_wq;

        
        struct usb_request *read_req;
        unsigned char *read_buf;
        unsigned read_count;
        unsigned bound;

        
        unsigned usermode_setup : 1,
                setup_in : 1,
                setup_out_ready : 1,
                setup_out_error : 1,
                setup_abort : 1;
        unsigned setup_wLength;

        wait_queue_head_t ctl_wq;

#define N_EVENT 5
        struct usb_gadgetfs_event event [N_EVENT];
        unsigned ev_next;

        unsigned char ctl_write_buf[CTL_WRITE_BUF_SZ];
        unsigned ctl_write_count;
        unsigned disconnect_pend;
};

static struct mtp_context _context;

#define LOG2_STATUS_POLL_MSEC   4

static struct usb_interface_descriptor intf_desc =
{
        .bLength =		sizeof intf_desc,
        .bDescriptorType =	USB_DT_INTERFACE,
        .bNumEndpoints =	3,
        .bInterfaceClass =	0xff,
        .bInterfaceSubClass =	0x00,
        .bInterfaceProtocol =	0x00,
};

static struct usb_endpoint_descriptor hs_bulk_in_desc =
{
        .bLength =		USB_DT_ENDPOINT_SIZE,
        .bDescriptorType =	USB_DT_ENDPOINT,
        .bEndpointAddress =	USB_DIR_IN,
        .bmAttributes =		USB_ENDPOINT_XFER_BULK,
        .wMaxPacketSize =	__constant_cpu_to_le16(512),
        .bInterval =		0,
};
static struct usb_endpoint_descriptor fs_bulk_in_desc =
{
        .bLength =		USB_DT_ENDPOINT_SIZE,
        .bDescriptorType =	USB_DT_ENDPOINT,
        .bEndpointAddress =	USB_DIR_IN,
        .bmAttributes =		USB_ENDPOINT_XFER_BULK,
        .wMaxPacketSize =	__constant_cpu_to_le16(64),
        .bInterval =		0,
};

static struct usb_endpoint_descriptor hs_bulk_out_desc =
{
        .bLength =		USB_DT_ENDPOINT_SIZE,
        .bDescriptorType =	USB_DT_ENDPOINT,
        .bEndpointAddress =	USB_DIR_OUT,
        .bmAttributes =		USB_ENDPOINT_XFER_BULK,
        .wMaxPacketSize =	__constant_cpu_to_le16(512),
        .bInterval =		0,
};

static struct usb_endpoint_descriptor fs_bulk_out_desc =
{
        .bLength =		USB_DT_ENDPOINT_SIZE,
        .bDescriptorType =	USB_DT_ENDPOINT,
        .bEndpointAddress =	USB_DIR_OUT,
        .bmAttributes =		USB_ENDPOINT_XFER_BULK,
        .wMaxPacketSize =	__constant_cpu_to_le16(64),
        .bInterval =		0,
};

static struct usb_endpoint_descriptor hs_status_desc =
{
        .bLength =		USB_DT_ENDPOINT_SIZE,
        .bDescriptorType =	USB_DT_ENDPOINT,
        .bEndpointAddress =	USB_DIR_IN,
        .bmAttributes =		USB_ENDPOINT_XFER_INT,
        .wMaxPacketSize =	__constant_cpu_to_le16(64),
        .bInterval =		LOG2_STATUS_POLL_MSEC + 3,
};

static struct usb_endpoint_descriptor fs_status_desc =
{
        .bLength =		USB_DT_ENDPOINT_SIZE,
        .bDescriptorType =	USB_DT_ENDPOINT,
        .bEndpointAddress =	USB_DIR_IN,
        .bmAttributes =		USB_ENDPOINT_XFER_INT,
        .wMaxPacketSize =	__constant_cpu_to_le16(64),
        .bInterval =		(1 << LOG2_STATUS_POLL_MSEC),
};

static struct usb_function usb_func_mtp;

static struct usb_gadgetfs_event *
next_event (struct mtp_context *ctxt, enum usb_gadgetfs_event_type type)
{
        struct usb_gadgetfs_event *event;
        unsigned i;

        switch (type)
        {

                
        case GADGETFS_DISCONNECT:
                
                
        case GADGETFS_CONNECT:
                ctxt->ev_next = 0;
                break;

        case GADGETFS_SETUP:      
        case GADGETFS_SUSPEND:    
                
                for (i = 0; i != ctxt->ev_next; i++)
                {
                        if (ctxt->event [i].type != type)
                                continue;

                        DBG_CTL("discard old event[%d] %d\n", i, type);

                        ctxt->ev_next--;
                        if (i == ctxt->ev_next)
                                break;

                        
                        memmove (&ctxt->event [i], &ctxt->event [i + 1],
                                 sizeof (struct usb_gadgetfs_event)
                                 * (ctxt->ev_next - i));
                }
                break;

        default:
                BUG ();
        }

        DBG_CTL("event[%d] = %d\n", ctxt->ev_next, type);

        event = &ctxt->event [ctxt->ev_next++];
        BUG_ON (ctxt->ev_next > N_EVENT);
        memset (event, 0, sizeof *event);
        event->type = type;

        wake_up(&ctxt->ctl_wq);

        return event;
}

static inline int _lock(atomic_t *excl)
{
        if (atomic_inc_return(excl) == 1)
        {
                return 0;
        } else
        {
                atomic_dec(excl);
                return -1;
        }
}

static inline void _unlock(atomic_t *excl)
{
        atomic_set(excl, 0);
}


static void req_put(struct mtp_context *ctxt, struct list_head *head, struct usb_request *req)
{
        unsigned long flags;

        spin_lock_irqsave(&ctxt->lock, flags);
        list_add_tail(&req->list, head);
        spin_unlock_irqrestore(&ctxt->lock, flags);
}


static struct usb_request *req_get(struct mtp_context *ctxt, struct list_head *head)
{
        unsigned long flags;
        struct usb_request *req;

        spin_lock_irqsave(&ctxt->lock, flags);
        if (list_empty(head))
        {
                req = 0;
        } else
        {
                req = list_first_entry(head, struct usb_request, list);
                list_del(&req->list);
        }
        spin_unlock_irqrestore(&ctxt->lock, flags);
        return req;
}

static void mtp_complete_in(struct usb_endpoint *ept, struct usb_request *req)
{
        struct mtp_context *ctxt = req->context;

        if (req->status != 0)
                ctxt->error = 1;

        req_put(ctxt, &ctxt->tx_idle, req);

        wake_up(&ctxt->write_wq);
}

static void mtp_complete_out(struct usb_endpoint *ept, struct usb_request *req)
{
        struct mtp_context *ctxt = req->context;

        if (req->status != 0)
        {
                ctxt->error = 1;
                req_put(ctxt, &ctxt->rx_idle, req);
        } else
        {
                req_put(ctxt, &ctxt->rx_done, req);
        }

        wake_up(&ctxt->read_wq);
}

static ssize_t mtp_read_bulk(struct file *fp, char __user *buf,
                             size_t count, loff_t *pos)
{
        struct mtp_context *ctxt = &_context;
        struct usb_request *req;
        int r = count, xfer;
        int ret;
#ifdef MTP_SUPPORT_CANCEL
        int cancel_discard_data = 0;
        char __user *reading_buf = buf;
        size_t reading_count = count;
        unsigned long flags;
#else
#define reading_count count
#define reading_buf buf
#endif

        if (_lock(&ctxt->read_excl))
                return -EBUSY;

        
        while (!(ctxt->online || ctxt->error))
        {
                DBG_CTL("mtp_read: waiting for online state\n");
                ret = wait_event_interruptible(ctxt->read_wq, (ctxt->online || ctxt->error));
                if (ret < 0)
                {
                        _unlock(&ctxt->read_excl);
                        return ret;
                }
        }

#ifdef MTP_SUPPORT_CANCEL
        while (ctxt->cancel_req_excl == 1)
        {
                DBG_CTL("mtp_read_bulk canceling...\n");

                wait_event_interruptible(ctxt->cancel_wq, (ctxt->cancel_req_excl == 0));

                DBG_CTL("mtp_read_bulk leave canceling block...\n");
        }
#endif

        DBG_IO("mtp_read(%d)\n", count);

        while (reading_count > 0)
        {
                if (ctxt->error)
                {
                        DBG_ERR("mtp_read: ctxt err\n");
                        r = -EIO;
                        break;
                }

                
                while ((req = req_get(ctxt, &ctxt->rx_idle)))
                {
requeue_req:
                        req->length = TXN_MAX;
                        ret = usb_ept_queue_xfer(ctxt->out, req);
                        if (ret < 0)
                        {
                                DBG_ERR("mtp_read: failed to queue req %p (%d)\n", req, ret);
                                r = -EIO;
                                ctxt->error = 1;
                                req_put(ctxt, &ctxt->rx_idle, req);
                                goto fail;
                        } else
                        {
                                DBG_IO("rx %p queue\n", req);
                        }
                }

#ifdef MTP_SUPPORT_CANCEL
                if (cancel_discard_data == 1)
                {
                        DBG_CTL("mtp_read_bulk reset read data...\n");
                        reading_count = count;
                        reading_buf = buf;
                        spin_lock_irqsave(&ctxt->cancel_lock, flags);
                        if (ctxt->cancel_req_excl == 0)
                        {
                                cancel_discard_data = 0;
                                DBG_CTL("mtp_read_bulk set cancel_discard_data 0, start read data...\n");
                        }
                        spin_unlock_irqrestore(&ctxt->cancel_lock, flags);
                } else
#endif

                
                if (ctxt->read_count > 0)
                {
                        xfer = (ctxt->read_count < reading_count) ? ctxt->read_count : reading_count;

                        DBG_IO("mtp_read: xfer=%d\n", xfer);

                        if (copy_to_user(reading_buf, ctxt->read_buf, xfer))
                        {
                                r = -EFAULT;
                                break;
                        }
                        ctxt->read_buf += xfer;
                        ctxt->read_count -= xfer;
                        reading_buf += xfer;
                        reading_count -= xfer;

                        
                        if (ctxt->read_count == 0)
                        {
                                DBG_IO("mtp_read: req done %p\n", ctxt->read_req);

                                req_put(ctxt, &ctxt->rx_idle, ctxt->read_req);
                                ctxt->read_req = 0;

                                r = xfer;
                                break;
                        }
                        continue;
                }

                
                req = 0;
                ret = wait_event_interruptible(ctxt->read_wq,
                                               ((req = req_get(ctxt, &ctxt->rx_done)) || ctxt->error));

                if (req != 0)
                {
                        
                        if (req->actual == 0)
                                goto requeue_req;

                        ctxt->read_req = req;
                        ctxt->read_count = req->actual;
                        ctxt->read_buf = req->buf;
                        DBG_IO("rx %p %d\n", req, req->actual);
                }

                if (ret < 0)
                {
                        DBG_ERR("mtp_read: wait event err %d\n", ret);

                        r = ret;
                        break;
                }
        }

fail:
        _unlock(&ctxt->read_excl);
        return r;
}

static ssize_t mtp_write_bulk(struct file *fp, const char __user *buf,
                              size_t count, loff_t *pos)
{
        struct mtp_context *ctxt = &_context;
        struct usb_request *req = 0;
        int r = count, xfer;
        int ret;

        DBG_IO("mtp_write(%d)\n", count);

        if (_lock(&ctxt->write_excl))
                return -EBUSY;

        while (count > 0)
        {
                if (ctxt->error)
                {
                        r = -EIO;
                        break;
                }

                
                req = 0;
                ret = wait_event_interruptible(ctxt->write_wq,
                                               ((req = req_get(ctxt, &ctxt->tx_idle)) || ctxt->error));

                if (ret < 0)
                {
                        r = ret;
                        break;
                }

                if (req != 0)
                {
                        xfer = count > TXN_MAX ? TXN_MAX : count;
                        if (copy_from_user(req->buf, buf, xfer))
                        {
                                r = -EFAULT;
                                break;
                        }

                        req->length = xfer;
                        ret = usb_ept_queue_xfer(ctxt->in, req);
                        if (ret < 0)
                        {
                                DBG_ERR("mtp_write: xfer error %d\n", ret);
                                ctxt->error = 1;
                                r = -EIO;
                                break;
                        }

                        buf += xfer;
                        count -= xfer;

                        
                        req = 0;
                }
        }


        if (req)
                req_put(ctxt, &ctxt->tx_idle, req);

        _unlock(&ctxt->write_excl);
        return r;
}

static ssize_t mtp_write_status(struct file *fp, const char __user *buf,
                                size_t count, loff_t *pos)
{
        return 0;
}

#ifdef MTP_SUPPORT_CANCEL
static long mtp_ioctl_ctl(struct file *file,
                          unsigned int cmd, unsigned long arg)
{
        struct mtp_context *ctxt = &_context;
        unsigned long flags;
        int ret = 0;

        DBG_CTL("mtp_ioctl_ctl start\n");

        if (!ctxt->online)
        {
                return -EIO;
        }

        if (_lock(&ctxt->ctl_write_excl))
                return -EBUSY;

        
        switch (cmd)
        {
        case USB_MTP_FUNC_IOC_GET_DEVICE_STATUS_SET:
        {
                DBG_CTL("%s, USB_MTP_FUNC_IOC_GET_DEVICE_STATUS_SET, arg=%ld\n", __func__, arg);
                if (arg == 0x2001)
                {
                        unsigned int responsecode;

                        
                        responsecode = PTP_RESPONSE_OK_CAP;

                        memcpy(ctxt->ctl_write_buf, (unsigned char *) &responsecode, sizeof(responsecode));

                        ctxt->ctl_write_count = 4;
                        DBG_CTL("%s, %d %d %d %d, arg=%ld\n", __func__,
                               ctxt->ctl_write_buf[0],
                               ctxt->ctl_write_buf[1],
                               ctxt->ctl_write_buf[2],
                               ctxt->ctl_write_buf[3],
                               arg);

                        spin_lock_irqsave(&ctxt->cancel_lock, flags);
                        ctxt->cancel_req_excl = 0;
                        spin_unlock_irqrestore(&ctxt->cancel_lock, flags);
                        wake_up(&ctxt->cancel_wq);
                        ret = 0;
                }
                goto ioctl_done;
        }
        default:
        {
                DBG_ERR("%s, arg=%ld ioctl not handled\n", __func__, arg);
        }
        break;
        }

        
        ret = -ENOIOCTLCMD;
ioctl_done:

        _unlock(&ctxt->ctl_write_excl);
        return ret;
}
#endif

static ssize_t mtp_read_ctl(struct file *fp, char __user *buf,
                            size_t count, loff_t *pos)
{
        struct mtp_context *ctxt = &_context;
        ssize_t retval = 0;
        size_t len = count;
        unsigned long flags;
        unsigned i, n;
        unsigned send_disconnect = 0;

        DBG_CTL("mtp_read_ctl start: %d\n", count);

        if (!ctxt->online)
        {
                DBG_CTL("mtp_read_ctl: waiting for online state\n");
                retval = wait_event_interruptible(ctxt->ctl_wq,
                        (ctxt->online || ctxt->error || ctxt->disconnect_pend));
                if (retval < 0)
                {
                        return -EIO;
                }
        }

        
        if (len < sizeof ctxt->event [0])
        {
                DBG_ERR("mtp_read_ctl - invalid len=%d\n", len);
                return -EINVAL;
        }

        if (!ctxt->bound)
        {
                DBG_ERR("mtp_read_ctl - not attached\n");
                return -EIO;
        }

        
        if (_lock(&ctxt->ctl_read_excl))
        {
                DBG_ERR("mtp_read_ctl error: busy\n");
                return -EBUSY;
        }

        DBG_CTL("mtp_read_ctl - o: %d e: %d\n", ctxt->online, ctxt->error);

        if (ctxt->disconnect_pend == 1)
        {
            DBG_ERR("mtp_read_ctl - disconnect pending, send GADGETFS_DISCONNECT\n");
            next_event(ctxt, GADGETFS_DISCONNECT);
            ctxt->disconnect_pend = 0;
            usb_function_enable(MTP_FUNCTION_NAME, 0);
            send_disconnect = 1;
        }

        len -= len % sizeof (struct usb_gadgetfs_event);

        
        while (ctxt->ev_next == 0)
        {
                DBG_CTL("mtp_read_ctl: waiting for event\n");
                retval = wait_event_interruptible(ctxt->ctl_wq, ctxt->ev_next);
                if (retval < 0)
                {
                        _unlock(&ctxt->ctl_read_excl);
                        return retval;
                }
        }

        DBG_CTL("mtp_read_ctl - event read\n");

        spin_lock_irqsave (&ctxt->lock, flags);

        n = len / sizeof (struct usb_gadgetfs_event);
        if (ctxt->ev_next < n)
                n = ctxt->ev_next;

        
        for (i = 0; i < n; i++)
        {
                if (ctxt->event [i].type == GADGETFS_SETUP)
                {
                        
                        n = i + 1;
                        break;
                }
        }
        spin_unlock_irqrestore (&ctxt->lock, flags);
        len = n * sizeof (struct usb_gadgetfs_event);
        if (copy_to_user (buf, &ctxt->event, len))
                retval = -EFAULT;
        else
                retval = len;
        if (len > 0)
        {
                
                spin_lock_irqsave (&ctxt->lock, flags);
                if (ctxt->ev_next > n)
                {
                        memmove(&ctxt->event[0], &ctxt->event[n],
                                sizeof (struct usb_gadgetfs_event)
                                * (ctxt->ev_next - n));
                }
                ctxt->ev_next -= n;
                spin_unlock_irqrestore (&ctxt->lock, flags);
        }

        _unlock(&ctxt->ctl_read_excl);

        DBG_CTL("mtp_read_ctl - complete\n");

        if (send_disconnect == 1)
        {
            DBG_CTL("%s - we do disconnect job since cable disconnect before mtp enable\n", __func__);
            ctxt->online = 0;
            ctxt->ctl_write_count = 0;
            _unlock(&ctxt->ctl_enable_excl);
            _unlock(&ctxt->ctl_open_excl);
            _unlock(&ctxt->open_excl);
        }

        return retval;
}

static ssize_t mtp_write_ctl(struct file *fp, const char __user *buf,
                             size_t count, loff_t *pos)
{
        struct mtp_context *ctxt = &_context;
        int ret, xfer;

        DBG_CTL("mtp_write_ctl(%d)\n", count);

        if (_lock(&ctxt->ctl_write_excl))
                return -EBUSY;

        xfer = count > CTL_WRITE_BUF_SZ ? CTL_WRITE_BUF_SZ : count;
        if (copy_from_user(ctxt->ctl_write_buf, buf, xfer))
        {

                DBG_ERR("mtp_write_ctl: copy fail!\n");
                ret = -EFAULT;
                ctxt->ctl_write_count = 0;
        } else
        {
                DBG_CTL("mtp_write_ctl: wrote %d\n", xfer);
                ret = ctxt->ctl_write_count = xfer;
        }

        _unlock(&ctxt->ctl_write_excl);
        return ret;
}

static int mtp_open_enable(struct inode *ip, struct file *fp)
{
        struct mtp_context *ctxt = &_context;

        DBG_INFO("+%s\n", __func__);
        if (_lock(&ctxt->ctl_enable_excl))
        {
                DBG_ERR("%s: fail to lock ctl_enable_excl!\n", __func__);
                return -EBUSY;
        }

        DBG_CTL("enabling mtp function\n");
        usb_function_enable(MTP_FUNCTION_NAME, 1);

        
        ctxt->error = 0;
        ctxt->cancel_req_excl = 0;

        return 0;
}

static int mtp_release_enable(struct inode *ip, struct file *fp)
{
        struct mtp_context *ctxt = &_context;

        DBG_INFO("-%s\n", __func__);
        usb_function_enable(MTP_FUNCTION_NAME, 0);

        ctxt->cancel_req_excl = 0;
        ctxt->disconnect_pend = 0;

        _unlock(&ctxt->ctl_enable_excl);
        _unlock(&ctxt->ctl_open_excl);
        _unlock(&ctxt->open_excl);
        _unlock(&ctxt->read_excl);
        _unlock(&ctxt->write_excl);

        wake_up(&ctxt->cancel_wq);
        return 0;
}

static int mtp_open_ctl(struct inode *ip, struct file *fp)
{
        struct mtp_context *ctxt = &_context;

        DBG_INFO("+%s\n", __func__);
        if (_lock(&ctxt->ctl_open_excl))
        {
                DBG_ERR("%s: fail to lock ctl_open_excl!\n", __func__);
                return -EBUSY;
        }

        
        ctxt->error = 0;

        return 0;
}

static int mtp_release_ctl(struct inode *ip, struct file *fp)
{
        struct mtp_context *ctxt = &_context;

        DBG_INFO("-%s\n", __func__);
        _unlock(&ctxt->ctl_open_excl);
        return 0;
}

static int mtp_open(struct inode *ip, struct file *fp)
{
        struct mtp_context *ctxt = &_context;

        DBG_INFO("+%s\n", __func__);
        if (_lock(&ctxt->open_excl))
        {
                DBG_ERR("%s: fail to lock open_excl!\n", __func__);
                return -EBUSY;
        }

        
        ctxt->error = 0;

        return 0;
}

static int mtp_release(struct inode *ip, struct file *fp)
{
        struct mtp_context *ctxt = &_context;

        DBG_INFO("-%s\n", __func__);
        _unlock(&ctxt->open_excl);
        return 0;
}

static struct file_operations mtp_fops_bulk =
{
        .owner =   THIS_MODULE,
        .read =    mtp_read_bulk,
        .write =   mtp_write_bulk,
        .open =    mtp_open,
        .release = mtp_release,
};
static struct file_operations mtp_fops_status =
{
        .owner =   THIS_MODULE,
        .write =   mtp_write_status,
        .open =    mtp_open,
        .release = mtp_release,
};

static struct file_operations mtp_fops_enable =
{
        .owner =   THIS_MODULE,
        .open =    mtp_open_enable,
        .release = mtp_release_enable,
};

static struct file_operations mtp_fops_ctl =
{
        .owner =   THIS_MODULE,
        .read =    mtp_read_ctl,
        .write =   mtp_write_ctl,
        .open =    mtp_open_ctl,
        .release = mtp_release_ctl,
#ifdef MTP_SUPPORT_CANCEL
        .unlocked_ioctl = mtp_ioctl_ctl,
#endif
};

static struct miscdevice mtp_device_bulk =
{
        .minor = MISC_DYNAMIC_MINOR,
        .name = "mtp_bulk",
        .fops = &mtp_fops_bulk,
};

static struct miscdevice mtp_device_status =
{
        .minor = MISC_DYNAMIC_MINOR,
        .name = "mtp_status",
        .fops = &mtp_fops_status,
};

static struct miscdevice mtp_device_enable =
{
        .minor = MISC_DYNAMIC_MINOR,
        .name = "android_mtp_enable",
        .fops = &mtp_fops_enable,
};

static struct miscdevice mtp_device_ctl =
{
        .minor = MISC_DYNAMIC_MINOR,
        .name = "mtp_ctl",
        .fops = &mtp_fops_ctl,
};

static void mtp_unbind(void *_ctxt)
{
        struct mtp_context *ctxt = _ctxt;
        struct usb_request *req;

        DBG_CTL("mtp_unbind()\n");

        if (!ctxt->bound)
                return;

        while ((req = req_get(ctxt, &ctxt->rx_idle)))
        {
                usb_ept_free_req(ctxt->out, req);
        }
        while ((req = req_get(ctxt, &ctxt->tx_idle)))
        {
                usb_ept_free_req(ctxt->in, req);
        }
        if (ctxt->in)
        {
                usb_ept_fifo_flush(ctxt->in);
                usb_ept_enable(ctxt->in,  0);
                usb_free_endpoint(ctxt->in);
        }
        if (ctxt->out)
        {
                usb_ept_fifo_flush(ctxt->out);
                usb_ept_enable(ctxt->out,  0);
                usb_free_endpoint(ctxt->out);
        }
        if (ctxt->status)
        {
                usb_ept_fifo_flush(ctxt->status);
                usb_ept_enable(ctxt->status,  0);
                usb_free_endpoint(ctxt->status);
        }

        ctxt->online = 0;
        ctxt->error = 1;

        
        next_event(ctxt, GADGETFS_DISCONNECT);

        
        wake_up(&ctxt->read_wq);
        ctxt->bound = 0;
}

static void mtp_bind(void *_ctxt)
{
        struct mtp_context *ctxt = _ctxt;
        struct usb_request *req;
        int n;

        intf_desc.bInterfaceNumber =
                usb_msm_get_next_ifc_number(&usb_func_mtp);

        ctxt->in = usb_alloc_endpoint(USB_DIR_IN);
        if (ctxt->in)
        {
                hs_bulk_in_desc.bEndpointAddress = USB_DIR_IN | ctxt->in->num;
                fs_bulk_in_desc.bEndpointAddress = USB_DIR_IN | ctxt->in->num;
        }

        ctxt->out = usb_alloc_endpoint(USB_DIR_OUT);
        if (ctxt->out)
        {
                hs_bulk_out_desc.bEndpointAddress = USB_DIR_OUT|ctxt->out->num;
                fs_bulk_out_desc.bEndpointAddress = USB_DIR_OUT|ctxt->out->num;
        }

        ctxt->status = usb_alloc_endpoint(USB_DIR_IN);
        if (ctxt->status)
        {
                hs_status_desc.bEndpointAddress = USB_DIR_IN|ctxt->status->num;
                fs_status_desc.bEndpointAddress = USB_DIR_IN|ctxt->status->num;
        }

        for (n = 0; n < RX_REQ_MAX; n++)
        {
                req = usb_ept_alloc_req(ctxt->out, TXN_MAX);
                if (req == 0)
                {
                        ctxt->bound = 1;
                        goto fail;
                }

                req->context = ctxt;
                req->complete = mtp_complete_out;
                req_put(ctxt, &ctxt->rx_idle, req);
        }

        for (n = 0; n < TX_REQ_MAX; n++)
        {
                req = usb_ept_alloc_req(ctxt->in, TXN_MAX);
                if (req == 0)
                {
                        ctxt->bound = 1;
                        goto fail;
                }

                req->context = ctxt;
                req->complete = mtp_complete_in;
                req_put(ctxt, &ctxt->tx_idle, req);
        }

        

        DBG_CTL("mtp_bind() allocated %d rx and %d tx requests\n",
               RX_REQ_MAX, TX_REQ_MAX);

        ctxt->bound = 1;
        return;

fail:
        DBG_ERR("mtp_bind() could not allocate requests\n");
        mtp_unbind(ctxt);
}

static void mtp_configure(int configured, void *_ctxt)
{
        struct mtp_context *ctxt = _ctxt;
        struct usb_request *req;
        struct usb_gadgetfs_event *event;

        DBG_CTL("mtp_configure() %d\n", configured);

        if (configured)
        {

                ctxt->online = 1;
                ctxt->error = 0;
                ctxt->disconnect_pend = 0;
                DBG_CTL("%s - disconnect pending 0\n", __func__);

                if (usb_msm_get_speed() == USB_SPEED_HIGH)
                {
                        usb_configure_endpoint(ctxt->in, &hs_bulk_in_desc);
                        usb_configure_endpoint(ctxt->out, &hs_bulk_out_desc);
                        usb_configure_endpoint(ctxt->status, &hs_status_desc);
                } else
                {
                        usb_configure_endpoint(ctxt->in, &fs_bulk_in_desc);
                        usb_configure_endpoint(ctxt->out, &fs_bulk_out_desc);
                        usb_configure_endpoint(ctxt->status, &fs_status_desc);
                }

                usb_ept_enable(ctxt->in,  1);
                usb_ept_enable(ctxt->out, 1);
                usb_ept_enable(ctxt->status, 1);

                
                ctxt->read_buf = 0;
                ctxt->read_count = 0;
                if (ctxt->read_req)
                {
                        req_put(ctxt, &ctxt->rx_idle, ctxt->read_req);
                        ctxt->read_req = 0;
                }

                
                while ((req = req_get(ctxt, &ctxt->rx_done)))
                        req_put(ctxt, &ctxt->rx_idle, req);

                
                event = next_event(ctxt, GADGETFS_CONNECT);
                event->u.speed = usb_msm_get_speed();

                ctxt->ctl_write_count = 0;

        } else
        {
                ctxt->online = 0;
                ctxt->ctl_write_count = 0;

#if 0
                
                next_event(ctxt, GADGETFS_DISCONNECT);
#endif
        }

        
        wake_up(&ctxt->read_wq);
}

static void mtp_disconnect(void *_ctxt)
{
        struct mtp_context *ctxt = _ctxt;

        ctxt->online = 0;
        ctxt->disconnect_pend = 1;
        DBG_CTL("%s - disconnect pending 1\n", __func__);

        
        next_event(ctxt, GADGETFS_DISCONNECT);
}


#define MSOS_FEATURE_INDEX_EXTENDED_CONFIG_DESC 0x0004
#define MSOS_FEATURE_INDEX_EXTENDED_PROP_DESC   0x0005


typedef struct usb_ms_extended_descriptor
{
        
        unsigned int dwLength;
        unsigned short bcdVersion;
        unsigned short wIndex;
        unsigned char  bCount;               
        unsigned char rsvd[7];

        
        unsigned char bInterfaceNumber;
        unsigned char bInterfaceIncrement;   
        unsigned char bClass[8];             
        unsigned char buSubClass[8];        
        unsigned char rsvd2[6];
} usb_ms_extended_descriptor;

static struct usb_ms_extended_descriptor ms_ext_desc =
{
        .dwLength                = sizeof (ms_ext_desc),
        .bcdVersion              = 0x0100,
        .wIndex                  = 0x04,
        .bCount                  = 1,
        .bInterfaceNumber        = 0,
        .bInterfaceIncrement     = 0x01,
        .bClass                  = "MTP"
};


static int copy_string_descriptor(char *string, char *buffer, int buflen)
{
        int length, i;

        if (string)
        {
                length = strlen(string);

                
                buffer[0] = (2 * length) + 2;

                
                if (buffer[0] > buflen)
                {
                        return -1;
                }

                
                buffer[1] = USB_DT_STRING;

                for (i = 0; i < length; i++)
                {
                        buffer[2 * i + 2] = string[i];
                        buffer[2 * i + 3] = 0;
                }
        }

        return buffer[0];
}

static int mtp_setup(struct usb_ctrlrequest* req, void* buf, int len, void *_ctxt)
{
        struct mtp_context *ctxt = (struct mtp_context *) _ctxt;
        int value;
#ifdef MTP_SUPPORT_CANCEL
        unsigned long flags;
#endif

        DBG_CTL("mtp_setup: begin, len = %d\n", req->wLength);

        
        if ((req->bRequestType & USB_TYPE_MASK) == USB_TYPE_STANDARD)
        {

                unsigned short type = req->wValue >> 8;
                unsigned short id = req->wValue & 0xFF;
                char *strbuf = buf;

                DBG_CTL("mtp_setup: standard req type: %x\n", type);

                if (type == USB_DT_STRING)
                {

                        if (id == STRING_MSFT_SECRET)
                        {

                                value = copy_string_descriptor(MSFT_SECRET, strbuf, len);
                                if (value < 0)
                                {
                                        DBG_ERR("MTP error: string desc failure!\n");
                                        return value;
                                }

                                
                                strbuf[value] = 0x0C;
                                strbuf[value + 1] = 0x00;

                                
                                value += 2;
                                strbuf[0] = value;

                                return value;
                        }
                }
        }
        
        else if ((req->bRequestType & USB_TYPE_MASK) == USB_TYPE_CLASS)
        {

                struct usb_gadgetfs_event *event;

                ctxt->setup_in = (req->bRequestType & USB_DIR_IN) ? 1 : 0;
                ctxt->setup_wLength = req->wLength;
                ctxt->setup_out_ready = 0;
                ctxt->setup_out_error = 0;
                value = -1;

                switch (req->bRequest)
                {

                case PTP_CANCEL_REQUEST:
                        DBG_CTL("PTP class request: cancel\n");
                        value = 0;
#ifdef MTP_SUPPORT_CANCEL
                        spin_lock_irqsave(&ctxt->cancel_lock, flags);
                        ctxt->cancel_req_excl = 1;
                        spin_unlock_irqrestore(&ctxt->cancel_lock, flags);
#endif
                        break;

                case PTP_RESET:
                        DBG_CTL("PTP class request: reset\n");
                        value = 0;
                        break;

                case PTP_GET_DEVICE_STATUS:

                        DBG_CTL("PTP class request: status (wc=%d)\n", ctxt->ctl_write_count);

                        
                        if (ctxt->ctl_write_count <= 0)
                        {
                                unsigned int responsecode;

                                
                                responsecode = (PTP_RESPONSE_BUSY_CAP);

                                value = sizeof(responsecode);
                                memcpy(buf, (unsigned char *) &responsecode, value);
                        } else
                        {
                                
                                value = ctxt->ctl_write_count > req->wLength ? req->wLength : ctxt->ctl_write_count;
                                memcpy(buf, ctxt->ctl_write_buf, value);
                                ctxt->ctl_write_count = 0;

                                DBG_CTL("PTP sending status resp: %x\n", *((unsigned int *) buf));

                                
                                return value;
                        }

                        break;

                default:
                        DBG_ERR("PTP class request: unknown (%d)\n", req->bRequest);
                        return -1;
                }

                DBG_CTL("mtp_setup: Send to userlevel req %04x\n", req->bRequest);
                event = next_event (ctxt, GADGETFS_SETUP);
                event->u.setup = *req;

                return value;
        }
        
        else if ((req->bRequestType & USB_TYPE_MASK) == USB_TYPE_VENDOR)
        {
                if (req->wIndex == MSOS_FEATURE_INDEX_EXTENDED_CONFIG_DESC)
                {

                        DBG_CTL("get ms descriptor: extended config (%d bytes)\n", req->wLength);

                        if (req->wLength > ms_ext_desc.dwLength)
                                value = ms_ext_desc.dwLength;
                        else
                                value = req->wLength;

                        memcpy(buf, &ms_ext_desc, value);

                        return value;

                } else if (req->wIndex == MSOS_FEATURE_INDEX_EXTENDED_PROP_DESC)
                {
                        
                        return -1;
                }
        }

        DBG_ERR("mtp_setup: not handled!\n");
        return -1;
}

static struct usb_function usb_func_mtp =
{
        .bind = mtp_bind,
        .unbind = mtp_unbind,
        .configure = mtp_configure,
        .disconnect = mtp_disconnect,
        .setup = mtp_setup,

        .name = MTP_FUNCTION_NAME,
        .context = &_context,
};

struct usb_descriptor_header *mtp_hs_descriptors[5];
struct usb_descriptor_header *mtp_fs_descriptors[5];

static int __init mtp_init(void)
{
        int ret = 0;

        struct mtp_context *ctxt = &_context;
        DBG_CTL("mtp_init()\n");

        ctxt->disconnect_pend = 0;
        DBG_CTL("%s - disconnect pending 0\n", __func__);

        init_waitqueue_head(&ctxt->read_wq);
        init_waitqueue_head(&ctxt->write_wq);
        init_waitqueue_head(&ctxt->ctl_wq);

        atomic_set(&ctxt->open_excl, 0);
        atomic_set(&ctxt->read_excl, 0);
        atomic_set(&ctxt->write_excl, 0);
        atomic_set(&ctxt->ctl_open_excl, 0);
        atomic_set(&ctxt->ctl_read_excl, 0);
        atomic_set(&ctxt->ctl_write_excl, 0);
        atomic_set(&ctxt->enable_excl, 0);

#ifdef MTP_SUPPORT_CANCEL
        init_waitqueue_head(&ctxt->cancel_wq);

        ctxt->cancel_req_excl = 0;

        spin_lock_init(&ctxt->cancel_lock);
#endif

        spin_lock_init(&ctxt->lock);

        INIT_LIST_HEAD(&ctxt->rx_idle);
        INIT_LIST_HEAD(&ctxt->rx_done);
        INIT_LIST_HEAD(&ctxt->tx_idle);

        mtp_hs_descriptors[0] = (struct usb_descriptor_header *)&intf_desc;
        mtp_hs_descriptors[1] =
                (struct usb_descriptor_header *)&hs_bulk_in_desc;
        mtp_hs_descriptors[2] =
                (struct usb_descriptor_header *)&hs_bulk_out_desc;
        mtp_hs_descriptors[3] =
                (struct usb_descriptor_header *)&hs_status_desc;
        mtp_hs_descriptors[4] = NULL;

        mtp_fs_descriptors[0] = (struct usb_descriptor_header *)&intf_desc;
        mtp_fs_descriptors[1] =
                (struct usb_descriptor_header *)&fs_bulk_in_desc;
        mtp_fs_descriptors[2] =
                (struct usb_descriptor_header *)&fs_bulk_out_desc;
        mtp_fs_descriptors[3] =
                (struct usb_descriptor_header *)&fs_status_desc;
        mtp_fs_descriptors[4] = NULL;

        usb_func_mtp.hs_descriptors = mtp_hs_descriptors;
        usb_func_mtp.fs_descriptors = mtp_fs_descriptors;

        ret = misc_register(&mtp_device_enable);
        if (ret)
        {
                DBG_ERR("mtp Can't register MTP enable device  %d \n",
                       MISC_DYNAMIC_MINOR);
                goto init_fail_01;
        }
        ret = misc_register(&mtp_device_bulk);
        if (ret)
        {
                DBG_ERR("mtp Can't register MTP bulk device  %d \n",
                       MISC_DYNAMIC_MINOR);
                goto init_fail_02;
        }
        ret = misc_register(&mtp_device_status);
        if (ret)
        {
                DBG_ERR("mtp Can't register MTP status device  %d \n",
                       MISC_DYNAMIC_MINOR);
                goto init_fail_03;
        }
        ret = misc_register(&mtp_device_ctl);
        if (ret)
        {
                DBG_ERR("mtp Can't register MTP control device  %d \n",
                       MISC_DYNAMIC_MINOR);
                goto init_fail_04;
        }

        ret = usb_function_register(&usb_func_mtp);
        return ret;

init_fail_04:
        misc_deregister(&mtp_device_ctl);
init_fail_03:
        misc_deregister(&mtp_device_status);
init_fail_02:
        misc_deregister(&mtp_device_bulk);
init_fail_01:
        misc_deregister(&mtp_device_enable);

        return ret;
}

module_init(mtp_init);

static void __exit mtp_exit(void)
{
        misc_deregister(&mtp_device_enable);
        misc_deregister(&mtp_device_bulk);
        misc_deregister(&mtp_device_status);
        misc_deregister(&mtp_device_ctl);

        usb_function_unregister(&usb_func_mtp);
}
module_exit(mtp_exit);
